#!/usr/bin/perl

use strict;
use warnings;
use utf8;

no warnings 'portable';

use Cwd;

my %types  = ();
my %dtrace = ();
my %allocs = ();

my $in;
my ($pointer, $ts);

my %frames = ();

my $kmod_dir = '/boot/kernel';

# Parse dtrace output.
while (my $line = <STDIN>) {
    chomp($line);
    next unless ($line);

    if ($line =~ /^([0-9]+) - MALLOC\+\+ (0x[^ ]+) type: (0x[^ ]+)/) {
        $in = 'malloc';
        $ts      = $1;
        $pointer = $2;
        my $type = $types{$3} || $3;

        unless (exists($dtrace{$pointer})) {
            $dtrace{$pointer} = {};
        }

        $dtrace{$pointer}{$ts} = {
            'malloc_type'  => $type,
            'malloc_stack' => []
        };
    } elsif ($line =~ /^([0-9]+) - MALLOC-- (0x[^ ]+) type: (0x[^ ]+)/) {
        $in = 'free';
        $ts      = $1;
        $pointer = $2;
        my $type = $types{$3} || $3;

        unless (exists($dtrace{$pointer})) {
            $dtrace{$pointer} = {};
        }

        $dtrace{$pointer}{$ts} = {
            'free_type'  => $type,
            'free_stack' => []
        };
    } elsif ($in && ($in eq 'malloc' || $in eq 'free')) {
        $line =~ s/^ +//;
        push($dtrace{$pointer}{$ts}{$in.'_stack'}, $line);
    } elsif ($line =~ /^malloc_type: ([^ ]+): +(0x[^ ]+)/) {
        $types{$2} = $1;
    }
}

# Sort dtrace records.
foreach $pointer (keys(%dtrace)) {
    $allocs{$pointer} = [{}];

    foreach $ts (sort(keys($dtrace{$pointer}))) {
        if (exists($dtrace{$pointer}{$ts}{'malloc_type'})) {
            if (exists($allocs{$pointer}[0]{'malloc_type'})) {
                unshift(@{$allocs{$pointer}}, $dtrace{$pointer}{$ts});
            } elsif (exists($allocs{$pointer}[0]{'free_type'})) {
                unshift(@{$allocs{$pointer}}, $dtrace{$pointer}{$ts});
            } else {
                $allocs{$pointer}[0]{'malloc_type'}  = $dtrace{$pointer}{$ts}{'malloc_type'};
                $allocs{$pointer}[0]{'malloc_stack'} = $dtrace{$pointer}{$ts}{'malloc_stack'};
            }
        } elsif (exists($dtrace{$pointer}{$ts}{'free_type'})) {
            if (exists($allocs{$pointer}[0]{'free_type'})) {
                unshift(@{$allocs{$pointer}}, $dtrace{$pointer}{$ts});
            } else {
                $allocs{$pointer}[0]{'free_type'}  = $dtrace{$pointer}{$ts}{'free_type'};
                $allocs{$pointer}[0]{'free_stack'} = $dtrace{$pointer}{$ts}{'free_stack'};
            }
        }
    }
}

sub print_stack ($)
{
    my ($stack) = @_;

    foreach my $frame (@$stack) {
        if (exists($frames{$frame})) {
            my $src = $frames{$frame};
            print "        $src\n";
        } elsif ($frame =~ /^([^`]+)`([^+]+)\+(0x.+)$/) {
            my $kmod = $1;
            my $func = $2;
            my $off  = hex($3);

            my $file = $kmod_dir.'/'.$kmod;

            # Compute address in kmod.
            my $addr = `nm $file | grep " $func\$" | cut -f1 -d ' '`;
            chomp($addr);
            $addr = hex($addr);
            $addr += $off;
            $addr = sprintf('0x%x', $addr);

            my $src = `addr2line -e $file $addr 2>/dev/null`;
            chomp($src);
            if ($src =~ /(.*):([0-9]+)$/) {
                my $src_file = Cwd::realpath($1) || $1;
                my $src_line = $2;
                $src = $src_file.':'.$src_line;
            }

            $src = $func."\n          ".$src;
            $frames{$frame} = $src;

            print "        $src\n";
        } else {
            print "        $frame\n";
        }
    }
}

my $errors_count = 0;

foreach $pointer (sort(keys(%allocs))) {
    my @instances = @{$allocs{$pointer}};

    for (my $i = $#instances; $i >= 0; $i--) {
        my %alloc = %{$instances[$i]};

        unless (exists($alloc{'malloc_type'})) {
            if ($i == 0) {
                print "--> Skip unmanaged pointer ($pointer):\n";
                print "      free(".$alloc{'free_type'}."):\n";
                print_stack($alloc{'free_stack'});
                print "\n";
            } else {
                my %previous = %{$instances[$i - 1]};
                print "--> Double-free ($pointer):\n";
                print "      mprevious(".$previous{'free_type'}."):\n";
                print_stack($previous{'free_stack'});
                print "      1st. free(".$previous{'free_type'}."):\n";
                print_stack($previous{'free_stack'});
                print "      2nd. free(".$alloc{'free_type'}."):\n";
                print_stack($alloc{'free_stack'});
                print "\n";
            }

            next;
        }

        unless (exists($alloc{'free_type'})) {
            print "--> Leaked allocation ($pointer):\n";
            print "      malloc(".$alloc{'malloc_type'}."):\n";
            print_stack($alloc{'malloc_stack'});
            print "\n";

            $errors_count++;
            next;
        }

        unless ($alloc{'malloc_type'} eq $alloc{'free_type'}) {
            print "--> Non-matching malloc_types ($pointer):\n";
            print "      malloc(".$alloc{'malloc_type'}."):\n";
            print_stack($alloc{'malloc_stack'});
            print "      free(".$alloc{'free_type'}."):\n";
            print_stack($alloc{'free_stack'});
            print "\n";

            $errors_count++;
            next;
        }
    }
}

if ($errors_count == 0) {
    print "No memory leak detected\n";
}
